
from argostranslate import package, translate
from packaging import version

import libretranslate.language


def boot(load_only=None, update_models=False, install_models=False):
    try:
        if update_models:
            check_and_install_models(load_only_lang_codes=load_only, update=update_models)
        else:
            check_and_install_models(force=install_models, load_only_lang_codes=load_only)
    except Exception as e:
        print("Cannot update models (normal if you're offline): %s" % str(e))


def check_and_install_models(force=False, load_only_lang_codes=None,update=False):
    if len(package.get_installed_packages()) < 2 or force or update:
        # Update package definitions from remote
        print("Updating language models")
        package.update_package_index()

        # Load available packages from local package index
        available_packages = package.get_available_packages()
        installed_packages = package.get_installed_packages()
        print("Found %s models" % len(available_packages))
        if load_only_lang_codes is not None:
            # load_only_lang_codes: List[str] (codes)
            load_only_lang_codes = libretranslate.language.iso2model(load_only_lang_codes)

            # Ensure the user does not use any unavailable language code.
            unavailable_lang_codes = set(load_only_lang_codes)
            for pack in available_packages:
                unavailable_lang_codes -= {pack.from_code, pack.to_code}
            if unavailable_lang_codes:
                raise ValueError(
                    "Unavailable language codes: %s."
                    % ",".join(sorted(unavailable_lang_codes))
                )
            # Keep only the packages that have both from_code and to_code in our list.
            available_packages = [
                pack
                for pack in available_packages
                if pack.from_code in load_only_lang_codes and pack.to_code in load_only_lang_codes
            ]
            if not available_packages:
                raise ValueError("no available package")
            print("Keep %s models" % len(available_packages))

        # Download and install all available packages
        for available_package in available_packages:
            update = False
            if not force:
                for pack in installed_packages:
                    if (
                            pack.from_code == available_package.from_code
                            and pack.to_code == available_package.to_code
                        ):
                        update = True
                        if version.parse(pack.package_version) < version.parse(available_package.package_version):
                            print(
                                f"Updating {available_package} ({pack.package_version}->{available_package.package_version}) ..."
                            )
                            pack.update()
            if not update:
                print(
                    f"Downloading {available_package} ({available_package.package_version}) ..."
                )
                available_package.install()

        # reload installed languages
        libretranslate.language.languages = translate.get_installed_languages()
        print(
            f"Loaded support for {len(translate.get_installed_languages())} languages ({len(available_packages)} models total)!"
        )



#!/usr/bin/env python
import os
import sys
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from libretranslate.main import get_args, get_parser

parser = get_parser()
args = get_args()
for arg in dir(args):
    if not arg.startswith('_'):
        value = getattr(args, arg)
        def_value = parser.get_default(arg)
        if not callable(value) and (value != def_value or arg == 'port' or arg == 'threads' or arg == 'host'):
            if isinstance(value, str):
                value = value.replace('"', '')
            elif isinstance(value, list):
                value = ",".join(value)
            print(f"export LT_{arg.upper()}={value}")



from libretranslate.storage import get_storage
import hashlib
import json
import gzip

cache = None
def get_translation_cache():
    return cache

class TranslationCache:
    def __init__(self, translation_cache_aks):
        self.enabled = len(translation_cache_aks) > 0
        self.api_keys = [ak for ak in translation_cache_aks if ak.lower() != "all"]
        self.cache_all = "all" in [ak.lower() for ak in translation_cache_aks]
        self.expire = 604800 # 7 days
        self.storage = get_storage()

        assert self.storage is not None, "Storage is none"

    def should_check(self, ak):
        return self.enabled and (self.cache_all or ak in self.api_keys)

    def hit(self, src_texts, source_lang, target_lang, text_format, num_alternatives):
        text_blob = "|".join(src_texts) if isinstance(src_texts, list) else src_texts
        fingerprint = f"{text_blob}:{source_lang}:{target_lang}:{text_format}:{num_alternatives}"
        cache_key = "tcache_" + hashlib.md5(fingerprint.encode('utf-8')).hexdigest()

        cached = self.storage.get_str(cache_key, raw=True)
        if len(cached) == 0:
            cached = None

        if cached is not None:
            try:
                cached = gzip.decompress(cached).decode('utf-8')
            except Exception as e:
                print(str(e))

        return cache_key, cached

    def cache(self, cache_key, content):
        try:
            if isinstance(content, dict):
                content = json.dumps(content)
                compressed = gzip.compress(content.encode('utf-8'))
                
            self.storage.set_str(cache_key, compressed, self.expire)
        except Exception as e:
            print(str(e))

def setup(translation_cache_aks):
    global cache
    
    cache = TranslationCache(translation_cache_aks)
    return cache


from libretranslate.storage import get_storage

active = False
threshold = -1

def forgive_banned():
    global threshold

    clear_list = []
    s = get_storage()
    banned = s.get_all_hash_int("banned")

    for ip in banned:
        if banned[ip] <= 0:
            clear_list.append(ip)
        else:
            s.set_hash_int("banned", ip, min(threshold, banned[ip]) - 1)

    for ip in clear_list:
        s.del_hash("banned", ip)

def setup(args):
    global active
    global threshold

    if args.req_flood_threshold > 0:
        active = True
        threshold = args.req_flood_threshold

def report(request_ip):
    if active:
        get_storage().inc_hash_int("banned", request_ip)

def decrease(request_ip):
    s = get_storage()
    if s.get_hash_int("banned", request_ip) > 0:
        s.dec_hash_int("banned", request_ip)

def has_violation(request_ip):
    s = get_storage()
    return s.get_hash_int("banned", request_ip) > 0

def is_banned(request_ip):
    s = get_storage()

    # More than X offences?
    return active and s.get_hash_int("banned", request_ip) >= threshold

def fingerprint_mismatch(request_ip, fingerprint):
    if not isinstance(fingerprint, str) or fingerprint == "":
        return True
    
    s = get_storage()
    k = f"fingerprint:{request_ip}"
    expected = s.get_str(k)
    if expected == "":
        s.set_str(k, fingerprint, ex=300)
        return False
    else:
        return fingerprint != expected


def test_api_get_health(client):
    response = client.get("/health")

    assert response.status_code == 200
    assert response.json == {"status": "ok"}


def test_api_health_must_fail_bad_request_type(client):
    response = client.post("/health")

    assert response.status_code == 405



import os
import sqlite3
import uuid

import requests
from expiringdict import ExpiringDict

from libretranslate.default_values import DEFAULT_ARGUMENTS as DEFARGS

DEFAULT_DB_PATH = DEFARGS['API_KEYS_DB_PATH']


class Database:
    def __init__(self, db_path=DEFAULT_DB_PATH, max_cache_len=1000, max_cache_age=30):
        # Legacy check - this can be removed at some point in the near future
        if os.path.isfile("api_keys.db") and not os.path.isfile("db/api_keys.db"):
            print("Migrating {} to {}".format("api_keys.db", "db/api_keys.db"))
            try:
                os.rename("api_keys.db", "db/api_keys.db")
            except Exception as e:
                print(str(e))

        db_dir = os.path.dirname(db_path)
        if db_dir != '' and not os.path.exists(db_dir):
            os.makedirs(db_dir)
        self.db_path = db_path
        self.cache = ExpiringDict(max_len=max_cache_len, max_age_seconds=max_cache_age)

        # Make sure to do data synchronization on writes!
        self.c = sqlite3.connect(db_path, check_same_thread=False)
        self.c.execute(
            """CREATE TABLE IF NOT EXISTS api_keys (
            "api_key"	TEXT NOT NULL,
            "req_limit"	INTEGER NOT NULL,
            "char_limit" INTEGER DEFAULT NULL,
            PRIMARY KEY("api_key")
        );"""
        )

        # Schema/upgrade checks
        schema = self.c.execute("SELECT sql FROM sqlite_master WHERE type='table' AND name='api_keys';").fetchone()[0]
        if '"char_limit" INTEGER DEFAULT NULL' not in schema:
            self.c.execute('ALTER TABLE api_keys ADD COLUMN "char_limit" INTEGER DEFAULT NULL;')

    def lookup(self, api_key):
        val = self.cache.get(api_key)
        if val is None:
            # DB Lookup
            stmt = self.c.execute(
                "SELECT req_limit, char_limit FROM api_keys WHERE api_key = ?", (api_key,)
            )
            row = stmt.fetchone()
            if row is not None:
                self.cache[api_key] = row
                val = row
            else:
                self.cache[api_key] = False
                val = False

        if isinstance(val, bool):
            val = None

        return val

    def add(self, req_limit, api_key="auto", char_limit=None):
        if api_key == "auto":
            api_key = str(uuid.uuid4())
        if char_limit == 0:
            char_limit = None

        self.remove(api_key)
        self.c.execute(
            "INSERT INTO api_keys (api_key, req_limit, char_limit) VALUES (?, ?, ?)",
            (api_key, req_limit, char_limit),
        )
        self.c.commit()
        return (api_key, req_limit, char_limit)

    def remove(self, api_key):
        self.c.execute("DELETE FROM api_keys WHERE api_key = ?", (api_key,))
        self.c.commit()
        return api_key

    def all(self):
        row = self.c.execute("SELECT api_key, req_limit, char_limit FROM api_keys")
        return row.fetchall()


class RemoteDatabase:
    def __init__(self, url, max_cache_len=1000, max_cache_age=600):
        self.url = url
        self.cache = ExpiringDict(max_len=max_cache_len, max_age_seconds=max_cache_age)

    def lookup(self, api_key):
        val = self.cache.get(api_key)
        if val is None:
            try:
                r = requests.post(self.url, data={'api_key': api_key}, timeout=60)
                res = r.json()
            except Exception as e:
                print("Cannot authenticate API key: " + str(e))
                return None

            if res.get('error') is not None:
                return None

            req_limit = res.get('req_limit', None)
            char_limit = res.get('char_limit', None)

            val = self.cache[api_key] = (req_limit, char_limit)
            
        return val



def test_api_get_spec(client):
    response = client.get("/spec")

    assert response.status_code == 200


def test_api_get_spec_must_fail_bad_request_type(client):
    response = client.post("/spec")

    assert response.status_code == 405



import os


class SuspiciousFileOperationError(Exception):
    pass


def path_traversal_check(unsafe_path, known_safe_path):
    known_safe_path = os.path.abspath(known_safe_path)
    unsafe_path = os.path.abspath(unsafe_path)

    if (os.path.commonprefix([known_safe_path, unsafe_path]) != known_safe_path):
        raise SuspiciousFileOperationError(f"{unsafe_path} is not safe")

    # Passes the check
    return unsafe_path



import argparse
import operator
import sys
import os

from libretranslate.app import create_app
from libretranslate.default_values import DEFAULT_ARGUMENTS as DEFARGS
from werkzeug.serving import run_simple
from werkzeug.middleware.dispatcher import DispatcherMiddleware

def get_parser():
    parser = argparse.ArgumentParser(
        description="LibreTranslate - Free and Open Source Translation API"
    )
    parser.add_argument(
        "--host", type=str, help="Hostname (%(default)s)", default=DEFARGS['HOST']
    )
    parser.add_argument("--port", type=int, help="Port (%(default)s)", default=DEFARGS['PORT'])
    parser.add_argument(
        "--char-limit",
        default=DEFARGS['CHAR_LIMIT'],
        type=int,
        metavar="<number of characters>",
        help="Set character limit (%(default)s)",
    )
    parser.add_argument(
        "--req-limit",
        default=DEFARGS['REQ_LIMIT'],
        type=int,
        metavar="<number>",
        help="Set the default maximum number of requests per minute per client (%(default)s)",
    )
    parser.add_argument(
        "--req-limit-storage",
        default=DEFARGS['REQ_LIMIT_STORAGE'],
        type=str,
        metavar="<Storage URI>",
        help="Storage URI to use for request limit data storage. See https://flask-limiter.readthedocs.io/en/stable/configuration.html. (%(default)s)",
    )
    parser.add_argument(
        "--hourly-req-limit",
        default=DEFARGS['HOURLY_REQ_LIMIT'],
        type=int,
        metavar="<number>",
        help="Set the default maximum number of requests per hour per client, in addition to req-limit. (%(default)s)",
    )
    parser.add_argument(
        "--hourly-req-limit-decay",
        default=DEFARGS['HOURLY_REQ_LIMIT_DECAY'],
        type=int,
        metavar="<number>",
        help="When used in combination with hourly-req-limit, adds additional hourly restrictions that logaritmically decrease for each additional hour. (%(default)s)",
    )
    parser.add_argument(
        "--daily-req-limit",
        default=DEFARGS['DAILY_REQ_LIMIT'],
        type=int,
        metavar="<number>",
        help="Set the default maximum number of requests per day per client, in addition to req-limit. (%(default)s)",
    )
    parser.add_argument(
        "--req-flood-threshold",
        default=DEFARGS['REQ_FLOOD_THRESHOLD'],
        type=int,
        metavar="<number>",
        help="Set the maximum number of request limit offences that a client can exceed before being banned. (%(default)s)",
    )
    parser.add_argument(
        "--req-time-cost",
        default=DEFARGS['REQ_TIME_COST'],
        type=int,
        metavar="<number>",
        help="Considers a time cost (in seconds) for request limiting purposes. If a request takes 10 seconds and this value is set to 5, the request cost is either 2 or the actual request cost (whichever is greater). (%(default)s)",
    )
    parser.add_argument(
        "--batch-limit",
        default=DEFARGS['BATCH_LIMIT'],
        type=int,
        metavar="<number of texts>",
        help="Set maximum number of texts to translate in a batch request (%(default)s)",
    )
    parser.add_argument(
        "--debug", default=DEFARGS['DEBUG'], action="store_true", help="Enable debug environment"
    )
    parser.add_argument(
        "--ssl", default=DEFARGS['SSL'], action="store_true", help="Whether to enable SSL"
    )
    parser.add_argument(
        "--frontend-language-source",
        type=str,
        default=DEFARGS['FRONTEND_LANGUAGE_SOURCE'],
        metavar="<language code>",
        help="Set frontend default language - source (%(default)s)",
    )
    parser.add_argument(
        "--frontend-language-target",
        type=str,
        default=DEFARGS['FRONTEND_LANGUAGE_TARGET'],
        metavar="<language code>",
        help="Set frontend default language - target (%(default)s)",
    )
    parser.add_argument(
        "--frontend-language",
        type=str,
        default=DEFARGS['FRONTEND_LANGUAGE'],
        metavar="<language code>",
        help="Set frontend interface language (%(default)s)",
    )
    parser.add_argument(
        "--frontend-title",
        type=str,
        default=DEFARGS['FRONTEND_TITLE'],
        metavar="<title>",
        help="Set frontend title (%(default)s)",
    )
    parser.add_argument(
        "--frontend-timeout",
        type=int,
        default=DEFARGS['FRONTEND_TIMEOUT'],
        metavar="<milliseconds>",
        help="Set frontend translation timeout (%(default)s)",
    )
    parser.add_argument(
        "--api-keys",
        default=DEFARGS['API_KEYS'],
        action="store_true",
        help="Enable API keys database for per-user rate limits lookup",
    )
    parser.add_argument(
        "--api-keys-db-path",
        default=DEFARGS['API_KEYS_DB_PATH'],
        type=str,
        help="Use a specific path inside the container for the local database. Can be absolute or relative (%(default)s)",
    )
    parser.add_argument(
        "--api-keys-remote",
        default=DEFARGS['API_KEYS_REMOTE'],
        type=str,
        help="Use this remote endpoint to query for valid API keys instead of using the local database",
    )
    parser.add_argument(
        "--get-api-key-link",
        default=DEFARGS['GET_API_KEY_LINK'],
        type=str,
        help="Show a link in the UI where to direct users to get an API key",
    )
    parser.add_argument(
        "--require-api-key-origin",
        type=str,
        default=DEFARGS['REQUIRE_API_KEY_ORIGIN'],
        help="Require use of an API key for programmatic access to the API, unless the request origin matches this domain",
    )
    parser.add_argument(
        "--require-api-key-secret",
        default=DEFARGS['REQUIRE_API_KEY_SECRET'],
        action="store_true",
        help="Require use of an API key for programmatic access to the API, unless the client also sends a secret match",
    )
    parser.add_argument(
        "--require-api-key-fingerprint",
        default=DEFARGS['REQUIRE_API_KEY_FINGERPRINT'],
        action="store_true",
        help="Require use of an API key for programmatic access to the API, unless the client also matches a fingerprint",
    )
    parser.add_argument(
        "--hide-api",
        default=DEFARGS['HIDE_API'],
        action="store_true",
        help="Hide the API request/response fields in the frontend",
    )
    parser.add_argument(
        "--under-attack",
        default=DEFARGS['UNDER_ATTACK'],
        action="store_true",
        help="Enable under attack mode. When enabled, requests must be made with an API key",
    )
    parser.add_argument(
        "--shared-storage",
        type=str,
        default=DEFARGS['SHARED_STORAGE'],
        metavar="<Storage URI>",
        help="Shared storage URI to use for multi-process data sharing (e.g. via gunicorn)",
    )
    parser.add_argument(
        "--secondary",
        default=DEFARGS['SECONDARY'],
        action="store_true",
        help="Mark this instance as a secondary instance to avoid conflicts with the primary node in multi-node setups",
    )
    parser.add_argument(
        "--load-only",
        type=operator.methodcaller("split", ","),
        default=DEFARGS['LOAD_ONLY'],
        metavar="<comma-separated language codes>",
        help="Set available languages (ar,de,en,es,fr,ga,hi,it,ja,ko,pt,ru,zh)",
    )
    parser.add_argument(
        "--alternatives-limit",
        default=DEFARGS['ALTERNATIVES_LIMIT'],
        type=int,
        metavar="<maximum number of alternatives translations>",
        help="Set the maximum number of supported alternative translations (%(default)s)",
    )
    parser.add_argument(
        "--threads",
        default=DEFARGS['THREADS'],
        type=int,
        metavar="<number of threads>",
        help="Set number of threads (%(default)s)",
    )
    parser.add_argument(
        "--suggestions", default=DEFARGS['SUGGESTIONS'], action="store_true", help="Allow user suggestions"
    )
    parser.add_argument(
        "--disable-files-translation", default=DEFARGS['DISABLE_FILES_TRANSLATION'], action="store_true",
        help="Disable files translation"
    )
    parser.add_argument(
        "--disable-web-ui", default=DEFARGS['DISABLE_WEB_UI'], action="store_true", help="Disable web ui"
    )
    parser.add_argument(
        "--update-models", default=DEFARGS['UPDATE_MODELS'], action="store_true", help="Update language models at startup"
    )
    parser.add_argument(
        "--force-update-models", default=DEFARGS['FORCE_UPDATE_MODELS'], action="store_true", help="Install/Reinstall language models at startup"
    )
    parser.add_argument(
        "--metrics",
        default=DEFARGS['METRICS'],
        action="store_true",
        help="Enable the /metrics endpoint for exporting Prometheus usage metrics",
    )
    parser.add_argument(
        "--metrics-auth-token",
        default=DEFARGS['METRICS_AUTH_TOKEN'],
        type=str,
        help="Protect the /metrics endpoint by allowing only clients that have a valid Authorization Bearer token (%(default)s)",
    )
    parser.add_argument(
        "--translation-cache",
        type=operator.methodcaller("split", ","),
        default=DEFARGS['TRANSLATION_CACHE'],
        metavar="<comma separated API keys or 'all'>",
        help="Cache translation output for users with a particular API key (or 'all' to cache all translations)",
    )
    parser.add_argument(
        "--url-prefix",
        default=DEFARGS['URL_PREFIX'],
        type=str,
        help="Add a prefix like /url-prefix to URL: example.com:5000/url-prefix/",
    )
    return parser

def get_args():
    parser = get_parser()
    args = parser.parse_args()
    if args.url_prefix and not args.url_prefix.startswith('/'):
        args.url_prefix = '/' + args.url_prefix
    return args


def main():
    args = get_args()

    if args.url_prefix:
        def redirect(environ, start_response):
            start_response("301 REDIRECT", [("Content-Type", "text/plain"), ("Location", args.url_prefix)])
            yield b"Redirect..."

        app = DispatcherMiddleware(redirect, {
            args.url_prefix: create_app(args)
        })
    else:
        app = DispatcherMiddleware(create_app(args))

    if '--wsgi' in sys.argv:
        return app
    else:
        if args.debug and args.host == "*":
            # '::' will listen on both ipv6 and ipv4
            args.host = "::"

        if args.debug:
            run_simple(args.host, args.port, app)
        else:
            from waitress import serve

            url_scheme = "https" if args.ssl else "http"
            print(f"Running on {url_scheme}://{args.host}:{args.port}{args.url_prefix}")

            serve(
                app,
                host=args.host,
                port=args.port,
                url_scheme=url_scheme,
                threads=args.threads
            )


if __name__ == "__main__":
    main()



import sys

import pytest

from libretranslate.app import create_app
from libretranslate.main import get_args


@pytest.fixture()
def app():
    sys.argv = ['', '--load-only', 'en,es']
    app = create_app(get_args())

    yield app


@pytest.fixture()
def client(app):
    return app.test_client()



import atexit

from apscheduler.schedulers.background import BackgroundScheduler

scheduler = None

def setup(args):
    from libretranslate.flood import forgive_banned
    from libretranslate.secret import rotate_secrets

    global scheduler

    if scheduler is None:
        scheduler = BackgroundScheduler()

        if not args.secondary and args.req_flood_threshold > 0:
            scheduler.add_job(func=forgive_banned, trigger="interval", minutes=10)

        if not args.secondary and args.api_keys and args.require_api_key_secret:
            scheduler.add_job(func=rotate_secrets, trigger="interval", minutes=30)

        scheduler.start()

        # Shut down the scheduler when exiting the app
        atexit.register(lambda: scheduler.shutdown())


import io
import math
import os
import re
import tempfile
import uuid
import sys
import warnings
from datetime import datetime
from functools import wraps
from html import unescape
from timeit import default_timer

import argostranslatefiles
from argostranslatefiles import get_supported_formats
from flask import Blueprint, Flask, Response, abort, jsonify, render_template, request, send_file, url_for, make_response
from flask_babel import Babel
from flask_swagger import swagger
from flask_swagger_ui import get_swaggerui_blueprint
from translatehtml import translate_html
from werkzeug.exceptions import HTTPException
from werkzeug.http import http_date
from werkzeug.utils import secure_filename

from libretranslate import flood, remove_translated_files, scheduler, secret, security, storage, cache
from libretranslate.language import model2iso, iso2model, detect_languages, improve_translation_formatting
from libretranslate.locales import (
    _,
    _lazy,
    get_alternate_locale_links,
    get_available_locale_codes,
    get_available_locales,
    gettext_escaped,
    gettext_html,
    lazy_swag,
)

from .api_keys import Database, RemoteDatabase
from .suggestions import Database as SuggestionsDatabase

# Ignore pytorch complaints
warnings.simplefilter(action='ignore', category=FutureWarning)

# Rough map of emoji characters
emojis = {e: True for e in \
  [ord(' ')] +                    # Spaces
  list(range(0x1F600,0x1F64F)) +  # Emoticons
  list(range(0x1F300,0x1F5FF)) +  # Misc Symbols and Pictographs
  list(range(0x1F680,0x1F6FF)) +  # Transport and Map
  list(range(0x2600,0x26FF)) +    # Misc symbols
  list(range(0x2700,0x27BF)) +    # Dingbats
  list(range(0xFE00,0xFE0F)) +    # Variation Selectors
  list(range(0x1F900,0x1F9FF)) +  # Supplemental Symbols and Pictographs
  list(range(0x1F1E6,0x1F1FF)) +  # Flags
  list(range(0x20D0,0x20FF))      # Combining Diacritical Marks for Symbols
}

def get_version():
    try:
        with open("VERSION") as f:
            return f.read().strip()
    except:
        return "?"


def get_upload_dir():
    upload_dir = os.path.join(tempfile.gettempdir(), "libretranslate-files-translate")

    if not os.path.isdir(upload_dir):
        os.mkdir(upload_dir)

    return upload_dir


def get_req_api_key():
    if request.is_json:
        json = get_json_dict(request)
        ak = json.get("api_key")
    else:
        ak = request.values.get("api_key")

    return ak

def get_req_secret():
    if request.is_json:
        json = get_json_dict(request)
        ak = json.get("secret")
    else:
        ak = request.values.get("secret")

    return ak


def get_json_dict(request):
    d = request.get_json()
    if not isinstance(d, dict):
        abort(400, description=_("Invalid JSON format"))
    return d


def get_remote_address():
    if request.headers.getlist("X-Forwarded-For"):
        ip = request.headers.getlist("X-Forwarded-For")[0].split(",")[0]
    else:
        ip = request.remote_addr or "127.0.0.1"

    return ip

def get_fingerprint():
    return request.headers.get("User-Agent", "") + request.headers.get("Cookie", "")


def get_req_limits(default_limit, api_keys_db, db_multiplier=1, multiplier=1):
    req_limit = default_limit

    if api_keys_db:
        api_key = get_req_api_key()

        if api_key:
            api_key_limits = api_keys_db.lookup(api_key)
            if api_key_limits is not None:
                req_limit = api_key_limits[0] * db_multiplier

    return int(req_limit * multiplier)


def get_char_limit(default_limit, api_keys_db):
    char_limit = default_limit

    if api_keys_db:
        api_key = get_req_api_key()

        if api_key:
            api_key_limits = api_keys_db.lookup(api_key)
            if api_key_limits is not None:
                if api_key_limits[1] is not None:
                    char_limit = api_key_limits[1]

    return char_limit


def get_routes_limits(args, api_keys_db):
    default_req_limit = args.req_limit
    if default_req_limit == -1:
        # TODO: better way?
        default_req_limit = 9999999999999

    def minute_limits():
        return "%s per minute" % get_req_limits(default_req_limit, api_keys_db)

    def hourly_limits(n):
        def func():
          decay = (0.75 ** (n - 1))
          return "{} per {} hour".format(get_req_limits(args.hourly_req_limit * n, api_keys_db, int(os.environ.get("LT_HOURLY_REQ_LIMIT_MULTIPLIER", 60) * n), decay), n)
        return func

    def daily_limits():
        return "%s per day" % get_req_limits(args.daily_req_limit, api_keys_db, int(os.environ.get("LT_DAILY_REQ_LIMIT_MULTIPLIER", 1440)))

    res = [minute_limits]

    if args.hourly_req_limit > 0:
      for n in range(1, args.hourly_req_limit_decay + 2):
        res.append(hourly_limits(n))

    if args.daily_req_limit > 0:
        res.append(daily_limits)

    return res

def filter_unique(seq, extra):
    seen = set({extra, ""})
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]


def detect_translatable(src_texts):
  if isinstance(src_texts, list):
    return any(detect_translatable(t) for t in src_texts)

  for ch in src_texts:
    if not (ord(ch) in emojis):
      return True

  # All emojis
  return False


def create_app(args):
    from libretranslate.init import boot

    boot(args.load_only, args.update_models, args.force_update_models)

    from libretranslate.language import load_languages

    swagger_url = args.url_prefix + "/docs"  # Swagger UI (w/o trailing '/')
    api_url = "/spec"

    bp = Blueprint('Main app', __name__)

    storage.setup(args.shared_storage)
    trans_cache = cache.setup(args.translation_cache)

    if not args.disable_files_translation:
        remove_translated_files.setup(get_upload_dir())
    languages = load_languages()
    language_pairs = {}
    for lang in languages:
        language_pairs[lang.code] = sorted([l.to_lang.code for l in lang.translations_from])

    # Map userdefined frontend languages to argos language object.
    if args.frontend_language_source == "auto":
        frontend_argos_language_source = type(
            "obj", (object,), {"code": "auto", "name": _("Auto Detect")}
        )
    else:
        frontend_argos_language_source = next(
            iter([l for l in languages if l.code == args.frontend_language_source]),
            None,
        )
    if frontend_argos_language_source is None:
        frontend_argos_language_source = languages[0]


    language_target_fallback = languages[1] if len(languages) >= 2 else languages[0]

    if args.frontend_language_target == "locale":
      def resolve_language_locale():
          loc = get_locale()
          language_target = next(
              iter([l for l in languages if l.code == loc]), None
          )
          if language_target is None:
            language_target = language_target_fallback
          return language_target

      frontend_argos_language_target = resolve_language_locale
    else:
      language_target = next(
          iter([l for l in languages if l.code == args.frontend_language_target]), None
      )
      if language_target is None:
        language_target = language_target_fallback
      frontend_argos_language_target = lambda: language_target

    frontend_argos_supported_files_format = []

    for file_format in get_supported_formats():
        for ff in file_format.supported_file_extensions:
            frontend_argos_supported_files_format.append(ff)

    api_keys_db = None

    if args.req_limit > 0 or args.api_keys or args.daily_req_limit > 0 or args.hourly_req_limit > 0:
        api_keys_db = None
        if args.api_keys:
            api_keys_db = RemoteDatabase(args.api_keys_remote) if args.api_keys_remote else Database(args.api_keys_db_path)

        from flask_limiter import Limiter

        def limits_cost():
          req_cost = getattr(request, 'req_cost', 1)
          if args.req_time_cost > 0:
            return max(req_cost, int(math.ceil(getattr(request, 'duration', 0) / args.req_time_cost)))
          else:
            return req_cost

        def get_limits_key_func():
          if args.api_keys:
            def func():
              ak = get_req_api_key()
              return ak if ak else get_remote_address()
            return func
          else:
            return get_remote_address

        limiter = Limiter(
            key_func=get_limits_key_func(),
            default_limits=get_routes_limits(
                args, api_keys_db
            ),
            storage_uri=args.req_limit_storage,
            default_limits_deduct_when=lambda req: True, # Force cost to be called after the request
            default_limits_cost=limits_cost,
            strategy="moving-window",
        )
    else:
        from .no_limiter import Limiter

        limiter = Limiter()

    if not "gunicorn" in os.environ.get("SERVER_SOFTWARE", ""):
      # Gunicorn starts the scheduler in the master process
      scheduler.setup(args)

    flood.setup(args)
    secret.setup(args)

    measure_request = None
    gauge_request = None
    if args.metrics:
      if os.environ.get("PROMETHEUS_MULTIPROC_DIR") is None:
          default_mp_dir = os.path.abspath(os.path.join("db", "prometheus"))
          if not os.path.isdir(default_mp_dir):
            os.mkdir(default_mp_dir)
          os.environ["PROMETHEUS_MULTIPROC_DIR"] = default_mp_dir

      from prometheus_client import CONTENT_TYPE_LATEST, CollectorRegistry, Gauge, Summary, generate_latest, multiprocess

      @bp.route("/metrics")
      @limiter.exempt
      def prometheus_metrics():
        if args.metrics_auth_token:
          authorization = request.headers.get('Authorization')
          if authorization != "Bearer " + args.metrics_auth_token:
            abort(401, description=_("Unauthorized"))

        registry = CollectorRegistry()
        multiprocess.MultiProcessCollector(registry)
        return Response(generate_latest(registry), mimetype=CONTENT_TYPE_LATEST)

      measure_request = Summary('libretranslate_http_request_duration_seconds', 'Time spent on request', ['endpoint', 'status', 'request_ip', 'api_key'])
      measure_request.labels('/translate', 200, '127.0.0.1', '')

      gauge_request = Gauge('libretranslate_http_requests_in_flight', 'Active requests', ['endpoint', 'request_ip', 'api_key'], multiprocess_mode='livesum')
      gauge_request.labels('/translate', '127.0.0.1', '')

    def access_check(f):
        @wraps(f)
        def func(*a, **kw):
            ip = get_remote_address()

            if flood.is_banned(ip):
                abort(403, description=_("Too many request limits violations"))

            if args.api_keys:
                ak = get_req_api_key()
                if ak and api_keys_db.lookup(ak) is None:
                    abort(
                        403,
                        description=_("Invalid API key"),
                    )
                else:
                  need_key = False
                  key_missing = api_keys_db.lookup(ak) is None

                  if (args.require_api_key_origin
                      and key_missing
                      and not re.match(args.require_api_key_origin, request.headers.get("Origin", ""))
                  ):
                    need_key = True

                  req_secret = get_req_secret()
                  if (args.require_api_key_secret
                    and key_missing
                    and not secret.secret_match(req_secret)
                  ):
                    need_key = True

                    if secret.secret_bogus_match(req_secret):
                      abort(make_response(jsonify({
                        'translatedText': secret.get_emoji(),
                        'alternatives': [],
                        'detectedLanguage': { 'confidence': 100, 'language': 'en' }
                      }), 200))

                  if (args.require_api_key_fingerprint
                    and key_missing):
                    if flood.fingerprint_mismatch(ip, get_fingerprint()):
                      need_key = True

                  if args.under_attack and key_missing:
                    need_key = True

                  if need_key:
                    description = _("Please contact the server operator to get an API key")
                    if args.get_api_key_link:
                        description = _("Visit %(url)s to get an API key", url=args.get_api_key_link)
                    abort(
                        400,
                        description=description,
                    )
                    flood.report(get_remote_address())
            return f(*a, **kw)

        if args.metrics:
          @wraps(func)
          def measure_func(*a, **kw):
              start_t = default_timer()
              status = 200
              ip = get_remote_address()
              ak = get_req_api_key() or ''
              g = gauge_request.labels(request.path, ip, ak)
              try:
                g.inc()
                return func(*a, **kw)
              except HTTPException as e:
                status = e.code
                if status == 403:
                  ak = '' # Don't record invalid API keys
                raise e
              finally:
                request.duration = max(default_timer() - start_t, 0)
                measure_request.labels(request.path, status, ip, ak).observe(request.duration)
                g.dec()
          return measure_func
        else:
          @wraps(func)
          def time_func(*a, **kw):
            start_t = default_timer()
            try:
              return func(*a, **kw)
            finally:
              request.duration = max(default_timer() - start_t, 0)
          return time_func

    @bp.errorhandler(400)
    def invalid_api(e):
        return jsonify({"error": str(e.description)}), 400

    @bp.errorhandler(500)
    def server_error(e):
        return jsonify({"error": str(e.description)}), 500

    @bp.errorhandler(429)
    def slow_down_error(e):
        flood.report(get_remote_address())
        return jsonify({"error": _("Slowdown:") + " " + str(e.description)}), 429

    @bp.errorhandler(403)
    def denied(e):
        return jsonify({"error": str(e.description)}), 403

    @bp.route("/")
    @limiter.exempt
    def index():
        if args.disable_web_ui:
            abort(404)

        langcode = args.frontend_language if len(args.frontend_language) > 0 else request.args.get('lang')
        lang_cookie = None
        if langcode and langcode in get_available_locale_codes(not args.debug):
            lang_cookie = langcode

        resp = make_response(render_template(
            "index.html",
            frontendTimeout=args.frontend_timeout,
            api_keys=args.api_keys,
            get_api_key_link=args.get_api_key_link,
            web_version=os.environ.get("LT_WEB") is not None,
            version=get_version(),
            swagger_url=swagger_url,
            available_locales=sorted([{'code': l['code'], 'name': _lazy(l['name'])} for l in get_available_locales(not args.debug)], key=lambda s: s['name']),
            current_locale=get_locale(lang_cookie),
            alternate_locales=get_alternate_locale_links(),
            under_attack=args.under_attack,
            hide_api=args.hide_api,
            frontend_title=args.frontend_title,
        ))

        if args.require_api_key_secret:
          resp.set_cookie('r', '1')

        if lang_cookie is not None:
          resp.set_cookie('preferred_lang', langcode)

        return resp

    @bp.route("/js/app.js")
    @limiter.exempt
    def appjs():
      if args.disable_web_ui:
        abort(404)

      api_secret = ""
      bogus_api_secret = ""
      if args.require_api_key_secret:
        bogus_api_secret = secret.get_bogus_secret_b64()

        if 'User-Agent' in request.headers and request.cookies.get('r'):
          api_secret = secret.get_current_secret_js()
        else:
          api_secret = secret.get_bogus_secret_js()

      response = Response(render_template("app.js.template",
            url_prefix=args.url_prefix,
            get_api_key_link=args.get_api_key_link,
            api_secret=api_secret,
            bogus_api_secret=bogus_api_secret,
            under_attack=args.under_attack), content_type='application/javascript; charset=utf-8')

      if args.require_api_key_secret:
        response.headers['Last-Modified'] = http_date(datetime.now())
        response.headers['Cache-Control'] = 'no-store, no-cache, must-revalidate, post-check=0, pre-check=0, max-age=0'
        response.headers['Pragma'] = 'no-cache'
        response.headers['Expires'] = '-1'

      return response

    @bp.get("/languages")
    @limiter.exempt
    def langs():
        """
        Get Supported Languages
        ---
        tags:
          - translate
        responses:
          200:
            description: List of supported languages
            schema:
              id: languages
              type: array
              items:
                type: object
                description: Supported language
                properties:
                  code:
                    type: string
                    description: Language code
                  name:
                    type: string
                    description: Human-readable language name (in English)
                  targets:
                    type: array
                    items:
                      type: string
                    description: Supported target language codes
        """
        return jsonify([{"code": model2iso(l.code),
                         "name": _lazy(l.name),
                         "targets": model2iso(language_pairs.get(l.code, []))
                        } for l in languages])

    @bp.get("/health")
    @limiter.exempt
    def health():
        """
        Health Check
        ---
        tags:
          - misc
        responses:
          200:
            description: Service is healthy
            schema:
              id: health-response
              type: object
              properties:
                status:
                  type: string
                  description: Health status
                  example: ok
        """
        return jsonify({"status": "ok"})

    # Add cors
    @bp.after_request
    def after_request(response):
        response.headers.add("Access-Control-Allow-Origin", "*")
        response.headers.add(
            "Access-Control-Allow-Headers", "Authorization, Content-Type"
        )
        response.headers.add("Access-Control-Expose-Headers", "Authorization")
        response.headers.add("Access-Control-Allow-Methods", "GET, POST")
        response.headers.add("Access-Control-Allow-Credentials", "true")
        response.headers.add("Access-Control-Max-Age", 60 * 60 * 24 * 20)
        return response

    @bp.post("/translate")
    @access_check
    def translate():
        """
        Translate Text
        ---
        tags:
          - translate
        parameters:
          - in: formData
            name: q
            schema:
              oneOf:
                - type: string
                  example: Hello world!
                - type: array
                  example: ['Hello world!']
            required: true
            description: Text(s) to translate
          - in: formData
            name: source
            schema:
              type: string
              example: en
            required: true
            description: Source language code or "auto" for auto detection
          - in: formData
            name: target
            schema:
              type: string
              example: es
            required: true
            description: Target language code
          - in: formData
            name: format
            schema:
              type: string
              enum: [text, html]
              default: text
              example: text
            required: false
            description: >
              Format of source text:
               * `text` - Plain text
               * `html` - HTML markup
          - in: formData
            name: alternatives
            schema:
              type: integer
              default: 0
              example: 3
            required: false
            description: Preferred number of alternative translations
          - in: formData
            name: api_key
            schema:
              type: string
              example: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
            required: false
            description: API key
        responses:
          200:
            description: Translation
            schema:
              id: translate
              type: object
              properties:
                translatedText:
                  oneOf:
                    - type: string
                    - type: array
                  description: Translated text(s)
                detectedLanguage:
                  oneOf:
                    - type: object
                      properties:
                        confidence:
                          type: number
                          format: float
                          minimum: 0
                          maximum: 100
                          description: Confidence value
                          example: 100
                        language:
                          type: string
                          description: Language code
                    - type: array
                      items:
                        type: object
                        properties:
                          confidence:
                            type: number
                            format: float
                            minimum: 0
                            maximum: 100
                            description: Confidence value
                            example: 100
                          language:
                            type: string
                            description: Language code
                alternatives:
                  oneOf:
                    - type: array
                      items:
                        type: string
                    - type: array
                      items:
                        type: array
                        items:
                          type: string
                  description: Alternative translations
              required:
                - translatedText
          400:
            description: Invalid request
            schema:
              id: error-response
              type: object
              properties:
                error:
                  type: string
                  description: Error message
          500:
            description: Translation error
            schema:
              id: error-response
              type: object
              properties:
                error:
                  type: string
                  description: Error message
          429:
            description: Slow down
            schema:
              id: error-slow-down
              type: object
              properties:
                error:
                  type: string
                  description: Reason for slow down
          403:
            description: Banned
            schema:
              id: error-response
              type: object
              properties:
                error:
                  type: string
                  description: Error message
        """
        if request.is_json:
            json = get_json_dict(request)
            q = json.get("q")
            source_lang = iso2model(json.get("source"))
            target_lang = iso2model(json.get("target"))
            text_format = json.get("format")
            num_alternatives = int(json.get("alternatives", 0))
        else:
            q = request.values.get("q")
            source_lang = iso2model(request.values.get("source"))
            target_lang = iso2model(request.values.get("target"))
            text_format = request.values.get("format")
            num_alternatives = request.values.get("alternatives", 0)

        if not q:
            abort(400, description=_("Invalid request: missing %(name)s parameter", name='q'))
        if not source_lang:
            abort(400, description=_("Invalid request: missing %(name)s parameter", name='source'))
        if not target_lang:
            abort(400, description=_("Invalid request: missing %(name)s parameter", name='target'))

        try:
            num_alternatives = max(0, int(num_alternatives))
        except ValueError:
            abort(400, description=_("Invalid request: %(name)s parameter is not a number", name='alternatives'))

        if args.alternatives_limit != -1 and num_alternatives > args.alternatives_limit:
            abort(400, description=_("Invalid request: %(name)s parameter must be <= %(value)s", name='alternatives', value=args.alternatives_limit))

        if not request.is_json:
            # Normalize line endings to UNIX style (LF) only so we can consistently
            # enforce character limits.
            # https://www.rfc-editor.org/rfc/rfc2046#section-4.1.1
            q = "\n".join(q.splitlines())

        char_limit = get_char_limit(args.char_limit, api_keys_db)

        batch = isinstance(q, list)

        if batch and args.batch_limit != -1:
            batch_size = len(q)
            if args.batch_limit < batch_size:
                abort(
                    400,
                    description=_("Invalid request: request (%(size)s) exceeds text limit (%(limit)s)", size=batch_size, limit=args.batch_limit),
                )

        src_texts = q if batch else [q]

        ak = get_req_api_key()
        cache_key = None
        if trans_cache.should_check(ak):
          cache_key, hit = trans_cache.hit(src_texts, source_lang, target_lang, text_format, num_alternatives)
          if hit is not None:
            return Response(hit, status=200, mimetype="application/json")

        if char_limit != -1:
            for text in src_texts:
                if len(text) > char_limit:
                    abort(
                        400,
                        description=_("Invalid request: request (%(size)s) exceeds text limit (%(limit)s)", size=len(text), limit=char_limit),
                    )

        if batch:
            request.req_cost = max(1, len(q))

        translatable = detect_translatable(src_texts)
        if translatable:
          if source_lang == "auto":
              candidate_langs = detect_languages(src_texts)
              detected_src_lang = candidate_langs[0]
          else:
              detected_src_lang = {"confidence": 100.0, "language": source_lang}
        else:
          detected_src_lang = {"confidence": 0.0, "language": "en"}

        src_lang = next(iter([l for l in languages if l.code == detected_src_lang["language"]]), None)

        if src_lang is None:
            abort(400, description=_("%(lang)s is not supported", lang=source_lang))

        tgt_lang = next(iter([l for l in languages if l.code == target_lang]), None)

        if tgt_lang is None:
            abort(400, description=_("%(lang)s is not supported",lang=target_lang))

        if not text_format:
            text_format = "text"

        if text_format not in ["text", "html"]:
            abort(400, description=_("%(format)s format is not supported", format=text_format))

        try:
            if batch:
                batch_results = []
                batch_alternatives = []
                for text in q:
                    translator = src_lang.get_translation(tgt_lang)
                    if translator is None:
                        abort(400, description=_("%(tname)s (%(tcode)s) is not available as a target language from %(sname)s (%(scode)s)", tname=_lazy(tgt_lang.name), tcode=tgt_lang.code, sname=_lazy(src_lang.name), scode=src_lang.code))

                    if translatable:
                      if text_format == "html":
                          translated_text = unescape(str(translate_html(translator, text)))
                          alternatives = [] # Not supported for html yet
                      else:
                          hypotheses = translator.hypotheses(text, num_alternatives + 1)
                          translated_text = unescape(improve_translation_formatting(text, hypotheses[0].value))
                          alternatives = filter_unique([unescape(improve_translation_formatting(text, hypotheses[i].value)) for i in range(1, len(hypotheses))], translated_text)
                    else:
                      translated_text = text # Cannot translate, send the original text back
                      alternatives = []

                    batch_results.append(translated_text)
                    batch_alternatives.append(alternatives)

                result = {"translatedText": batch_results}

                if source_lang == "auto":
                    result["detectedLanguage"] = [model2iso(detected_src_lang)] * len(q)
                if num_alternatives > 0:
                    result["alternatives"] = batch_alternatives
            else:
                translator = src_lang.get_translation(tgt_lang)
                if translator is None:
                    abort(400, description=_("%(tname)s (%(tcode)s) is not available as a target language from %(sname)s (%(scode)s)", tname=_lazy(tgt_lang.name), tcode=tgt_lang.code, sname=_lazy(src_lang.name), scode=src_lang.code))

                if translatable:
                  if text_format == "html":
                      translated_text = unescape(str(translate_html(translator, q)))
                      alternatives = [] # Not supported for html yet
                  else:
                      hypotheses = translator.hypotheses(q, num_alternatives + 1)
                      translated_text = unescape(improve_translation_formatting(q, hypotheses[0].value))
                      alternatives = filter_unique([unescape(improve_translation_formatting(q, hypotheses[i].value)) for i in range(1, len(hypotheses))], translated_text)
                else:
                  translated_text = q # Cannot translate, send the original text back
                  alternatives = []

                result = {"translatedText": translated_text}

                if source_lang == "auto":
                    result["detectedLanguage"] = model2iso(detected_src_lang)
                if num_alternatives > 0:
                    result["alternatives"] = alternatives
            
            if cache_key is not None:
              trans_cache.cache(cache_key, result)

            return jsonify(result)
        except Exception as e:
            raise e
            abort(500, description=_("Cannot translate text: %(text)s", text=str(e)))

    @bp.post("/translate_file")
    @access_check
    def translate_file():
        """
        Translate a File
        ---
        tags:
          - translate
        consumes:
         - multipart/form-data
        parameters:
          - in: formData
            name: file
            type: file
            required: true
            description: File to translate
          - in: formData
            name: source
            schema:
              type: string
              example: en
            required: true
            description: Source language code  or "auto" for auto detection
          - in: formData
            name: target
            schema:
              type: string
              example: es
            required: true
            description: Target language code
          - in: formData
            name: api_key
            schema:
              type: string
              example: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
            required: false
            description: API key
        responses:
          200:
            description: Translated file
            schema:
              id: translate-file
              type: object
              properties:
                translatedFileUrl:
                  type: string
                  description: Translated file url
          400:
            description: Invalid request
            schema:
              id: error-response
              type: object
              properties:
                error:
                  type: string
                  description: Error message
          500:
            description: Translation error
            schema:
              id: error-response
              type: object
              properties:
                error:
                  type: string
                  description: Error message
          429:
            description: Slow down
            schema:
              id: error-slow-down
              type: object
              properties:
                error:
                  type: string
                  description: Reason for slow down
          403:
            description: Banned
            schema:
              id: error-response
              type: object
              properties:
                error:
                  type: string
                  description: Error message
        """
        if args.disable_files_translation:
            abort(403, description=_("Files translation are disabled on this server."))

        source_lang = iso2model(request.form.get("source"))
        target_lang = iso2model(request.form.get("target"))
        file = request.files['file']
        char_limit = get_char_limit(args.char_limit, api_keys_db)

        if not file:
            abort(400, description=_("Invalid request: missing %(name)s parameter", name='file'))
        if not source_lang:
            abort(400, description=_("Invalid request: missing %(name)s parameter", name='source'))
        if not target_lang:
            abort(400, description=_("Invalid request: missing %(name)s parameter", name='target'))

        if file.filename == '':
            abort(400, description=_("Invalid request: empty file"))

        if os.path.splitext(file.filename)[1] not in frontend_argos_supported_files_format:
            abort(400, description=_("Invalid request: file format not supported"))

        src_lang = next(iter([l for l in languages if l.code == source_lang]), None)

        if src_lang is None and source_lang != "auto":
            abort(400, description=_("%(lang)s is not supported", lang=source_lang))

        tgt_lang = next(iter([l for l in languages if l.code == target_lang]), None)

        if tgt_lang is None:
            abort(400, description=_("%(lang)s is not supported", lang=target_lang))

        try:
            filename = str(uuid.uuid4()) + '.' + secure_filename(file.filename)
            filepath = os.path.join(get_upload_dir(), filename)

            file.save(filepath)

            # Not an exact science: take the number of bytes and divide by
            # the character limit. Assuming a plain text file, this will
            # set the cost of the request to N = bytes / char_limit, which is
            # roughly equivalent to a batch process of N batches assuming
            # each batch uses all available limits
            if char_limit > 0:
                request.req_cost = max(1, int(os.path.getsize(filepath) / char_limit))

            if source_lang == "auto":
                src_texts = argostranslatefiles.get_texts(filepath)
                candidate_langs = detect_languages(src_texts)
                detected_src_lang = candidate_langs[0]
                src_lang = next(iter([l for l in languages if l.code == detected_src_lang["language"]]), None)
                if src_lang is None:
                    abort(400, description=_("%(lang)s is not supported", lang=detected_src_lang["language"]))

            translated_file_path = argostranslatefiles.translate_file(src_lang.get_translation(tgt_lang), filepath)
            translated_filename = os.path.basename(translated_file_path)

            return jsonify(
                {
                    "translatedFileUrl": url_for('Main app.download_file', filename=translated_filename, _external=True)
                }
            )
        except Exception as e:
            abort(500, description=e)

    @bp.get("/download_file/<string:filename>")
    def download_file(filename: str):
        """
        Download a translated file
        """
        if args.disable_files_translation:
            abort(400, description=_("Files translation are disabled on this server."))

        filepath = os.path.join(get_upload_dir(), filename)
        try:
            checked_filepath = security.path_traversal_check(filepath, get_upload_dir())
            if os.path.isfile(checked_filepath):
                filepath = checked_filepath
        except security.SuspiciousFileOperationError:
            abort(400, description=_("Invalid filename"))

        return_data = io.BytesIO()
        with open(filepath, 'rb') as fo:
            return_data.write(fo.read())
        return_data.seek(0)

        download_filename = filename.split('.')
        download_filename.pop(0)
        download_filename = '.'.join(download_filename)

        return send_file(return_data, as_attachment=True, download_name=download_filename)

    @bp.post("/detect")
    @access_check
    def detect():
        """
        Detect Language of Text
        ---
        tags:
          - translate
        parameters:
          - in: formData
            name: q
            schema:
              type: string
              example: What language is this?
            required: true
            description: Text to detect
          - in: formData
            name: api_key
            schema:
              type: string
              example: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
            required: false
            description: API key
        responses:
          200:
            description: Detections
            schema:
              id: detections
              type: array
              items:
                type: object
                properties:
                  confidence:
                    type: number
                    format: float
                    minimum: 0
                    maximum: 100
                    description: Confidence value
                    example: 100
                  language:
                    type: string
                    description: Language code
                    example: en
          400:
            description: Invalid request
            schema:
              id: error-response
              type: object
              properties:
                error:
                  type: string
                  description: Error message
          500:
            description: Detection error
            schema:
              id: error-response
              type: object
              properties:
                error:
                  type: string
                  description: Error message
          429:
            description: Slow down
            schema:
              id: error-slow-down
              type: object
              properties:
                error:
                  type: string
                  description: Reason for slow down
          403:
            description: Banned
            schema:
              id: error-response
              type: object
              properties:
                error:
                  type: string
                  description: Error message
        """
        if request.is_json:
            json = get_json_dict(request)
            q = json.get("q")
        else:
            q = request.values.get("q")

        if not q:
            abort(400, description=_("Invalid request: missing %(name)s parameter", name='q'))

        return jsonify(model2iso(detect_languages(q)))

    @bp.route("/frontend/settings")
    @limiter.exempt
    def frontend_settings():
        """
        Retrieve Frontend Settings
        ---
        tags:
          - misc
        responses:
          200:
            description: frontend settings
            schema:
              id: frontend-settings
              type: object
              properties:
                charLimit:
                  type: integer
                  description: Character input limit for this language (-1 indicates no limit)
                frontendTimeout:
                  type: integer
                  description: Frontend translation timeout
                apiKeys:
                  type: boolean
                  description: Whether the API key database is enabled.
                keyRequired:
                  type: boolean
                  description: Whether an API key is required.
                suggestions:
                  type: boolean
                  description: Whether submitting suggestions is enabled.
                supportedFilesFormat:
                  type: array
                  items:
                    type: string
                  description: Supported files format
                language:
                  type: object
                  properties:
                    source:
                      type: object
                      properties:
                        code:
                          type: string
                          description: Language code
                        name:
                          type: string
                          description: Human-readable language name (in English)
                    target:
                      type: object
                      properties:
                        code:
                          type: string
                          description: Language code
                        name:
                          type: string
                          description: Human-readable language name (in English)
        """
        target_lang = frontend_argos_language_target()

        return jsonify(
            {
                "charLimit": args.char_limit,
                "frontendTimeout": args.frontend_timeout,
                "apiKeys": args.api_keys,
                "keyRequired": bool(args.api_keys and args.require_api_key_origin),
                "suggestions": args.suggestions,
                "filesTranslation": not args.disable_files_translation,
                "supportedFilesFormat": [] if args.disable_files_translation else frontend_argos_supported_files_format,
                "language": {
                    "source": {
                        "code": frontend_argos_language_source.code,
                        "name": _lazy(frontend_argos_language_source.name),
                    },
                    "target": {
                        "code": target_lang.code,
                        "name": _lazy(target_lang.name),
                    },
                },
            }
        )

    @bp.post("/suggest")
    def suggest():
        """
        Submit a Suggestion to Improve a Translation
        ---
        tags:
          - misc
        parameters:
          - in: formData
            name: q
            schema:
              type: string
              example: Hello world!
            required: true
            description: Original text
          - in: formData
            name: s
            schema:
              type: string
              example: ¡Hola mundo!
            required: true
            description: Suggested translation
          - in: formData
            name: source
            schema:
              type: string
              example: en
            required: true
            description: Language of original text
          - in: formData
            name: target
            schema:
              type: string
              example: es
            required: true
            description: Language of suggested translation
        responses:
          200:
            description: Success
            schema:
              id: suggest-response
              type: object
              properties:
                success:
                  type: boolean
                  description: Whether submission was successful
          403:
            description: Not authorized
            schema:
              id: error-response
              type: object
              properties:
                error:
                  type: string
                  description: Error message
        """
        if not args.suggestions:
            abort(403, description=_("Suggestions are disabled on this server."))

        if request.is_json:
            json = get_json_dict(request)
            q = json.get("q")
            s = json.get("s")
            source_lang = iso2model(json.get("source"))
            target_lang = iso2model(json.get("target"))
        else:
            q = request.values.get("q")
            s = request.values.get("s")
            source_lang = iso2model(request.values.get("source"))
            target_lang = iso2model(request.values.get("target"))

        if not q:
            abort(400, description=_("Invalid request: missing %(name)s parameter", name='q'))
        if not s:
            abort(400, description=_("Invalid request: missing %(name)s parameter", name='s'))
        if not source_lang:
            abort(400, description=_("Invalid request: missing %(name)s parameter", name='source'))
        if not target_lang:
            abort(400, description=_("Invalid request: missing %(name)s parameter", name='target'))

        SuggestionsDatabase().add(q, s, source_lang, target_lang)
        return jsonify({"success": True})

    app = Flask(__name__)
    app.config["JSON_AS_ASCII"] = False

    if args.debug:
        app.config["TEMPLATES_AUTO_RELOAD"] = True

    app.register_blueprint(bp)

    limiter.init_app(app)

    swag = swagger(app)
    swag["basePath"] = args.url_prefix if args.url_prefix != "" else "/"
    swag["info"]["version"] = get_version()
    swag["info"]["title"] = "LibreTranslate"
    swag["info"]["description"] = "Free and Open Source Machine Translation API."
    swag["info"]["license"] = {"name": "AGPL-3.0"}

    @app.route(api_url)
    @limiter.exempt
    def spec():
        return jsonify(lazy_swag(swag))

    app.config["BABEL_TRANSLATION_DIRECTORIES"] = 'locales'

    def get_locale(cookie=None):
        available_codes = get_available_locale_codes()
        override_lang = request.headers.get('X-Override-Accept-Language')
        if override_lang and override_lang in available_codes:
            return override_lang
        qlang = request.args.get('lang')
        if qlang and qlang in available_codes:
          return qlang
        if cookie is None:
          cookie = request.cookies.get('preferred_lang')
        if cookie and cookie in available_codes:
          return cookie
        else:
          return request.accept_languages.best_match(get_available_locale_codes())

    Babel(app, locale_selector=get_locale)

    app.jinja_env.globals.update(_e=gettext_escaped, _h=gettext_html)

    # Call factory function to create our blueprint
    # The Blueprint is not using url_for which means the middleware does not work properly and we need to manually fix things
    swaggerui_blueprint = get_swaggerui_blueprint(swagger_url, args.url_prefix + api_url)
    swaggerui_blueprint.url_prefix = "/docs"
    app.register_blueprint(swaggerui_blueprint)

    if os.environ.get("LT_POWERCYCLE") is not None:
      print("Power cycling...")
      sys.exit(0)

    return app



import json


def test_api_translate(client):
    response = client.post("/translate", data={
        "q": "Hello",
        "source": "en",
        "target": "es",
        "format": "text"
    })

    response_json = json.loads(response.data)

    assert "translatedText" in response_json
    assert response.status_code == 200


def test_api_translate_batch(client):

    response = client.post("/translate", json={
        "q": ["Hello", "World"],
        "source": "en",
        "target": "es",
        "format": "text"
    })

    response_json = json.loads(response.data)

    assert "translatedText" in response_json
    assert isinstance(response_json["translatedText"], list)
    assert len(response_json["translatedText"]) == 2
    assert response.status_code == 200


def test_api_translate_unsupported_language(client):
    response = client.post("/translate", data={
        "q": "Hello",
        "source": "en",
        "target": "zz",
        "format": "text"
    })

    response_json = json.loads(response.data)

    assert "error" in response_json
    assert response_json["error"] == "zz is not supported"
    assert response.status_code == 400


def test_api_translate_missing_parameter(client):
    response = client.post("/translate", data={
        "source": "en",
        "target": "es",
        "format": "text"
    })

    response_json = json.loads(response.data)

    assert "error" in response_json
    assert response_json["error"] == "Invalid request: missing q parameter"
    assert response.status_code == 400



#!/usr/bin/env python
import os
import sys

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
import argparse

from libretranslate.init import check_and_install_models

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--load_only_lang_codes", type=str, default="")
    parser.add_argument("--update", action='store_true')
    args = parser.parse_args()
    lang_codes = args.load_only_lang_codes.split(",")
    if len(lang_codes) == 0 or lang_codes[0] == '':
        lang_codes = None
    if args.update:
        check_and_install_models(update=True, load_only_lang_codes=lang_codes)
    else:
        check_and_install_models(force=True, load_only_lang_codes=lang_codes)



#!/usr/bin/env python
import os
import sys

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
from babel.messages.frontend import main as pybabel

if __name__ == "__main__":
    if len(sys.argv) >= 2 and sys.argv[1] == 'mdtable':
        from libretranslate.locales import get_available_locales
        locales = get_available_locales(only_reviewed=False, sort_by_name=True)
        print("Language | Reviewed | Weblate Link")
        print("-------- | -------- | ------------")

        for l in locales:
            link = "https://hosted.weblate.org/translate/libretranslate/app/%s/" % l['code']
            if l['code'] == 'en':
                link = "https://hosted.weblate.org/projects/libretranslate/app/"
            print("{} | {} | {}".format(l['name'], ':heavy_check_mark:' if l['reviewed'] else '', "[Edit](%s)" % link))
    else:
        locales_dir = os.path.join("libretranslate", "locales")
        if not os.path.isdir(locales_dir):
            os.makedirs(locales_dir)

        print("Compiling locales")
        sys.argv = ["", "compile", "-f", "-d", locales_dir]
        pybabel()






import requests
import os
import sys

if os.path.isfile('/tmp/booting.flag'):
    sys.exit(0)

port = os.environ.get('LT_PORT', '5000')
response = requests.get(
    url=f'http://localhost:{port}/health',
    headers={'Content-Type': 'application/json'},
    json={},
    timeout=60
)
response.raise_for_status() # if server unavailable then requests with raise exception and healthcheck will fail



import json


def test_api_detect_language(client):
    response = client.post("/detect", data={
        "q": "Hello"
    })
    response_json = json.loads(response.data)

    assert "confidence" in response_json[0] and "language" in response_json[0]
    assert len(response_json) >= 1
    assert response.status_code == 200


def test_api_detect_language_must_fail_without_parameters(client):
    response = client.post("/detect")
    response_json = json.loads(response.data)

    assert "error" in response_json
    assert response.status_code == 400


def test_api_detect_language_must_fail_bad_request_type(client):
    response = client.get("/detect")

    assert response.status_code == 405



from libretranslate import main


def app(*args, **kwargs):
    import sys
    sys.argv = ['--wsgi']

    for k in kwargs:
        ck = k.replace("_", "-")
        if isinstance(kwargs[k], bool):
            if kwargs[k]:
                sys.argv.append("--" + ck)
        else:
            sys.argv.append("--" + ck)
            sys.argv.append(kwargs[k])

    return main()



import re
import sys

from prometheus_client import multiprocess


def child_exit(server, worker):
    multiprocess.mark_process_dead(worker.pid)

def on_starting(server):
    # Parse command line arguments
    proc_name = server.cfg.default_proc_name
    kwargs = {}
    if proc_name.startswith("wsgi:app"):
        str_args = re.sub(r'wsgi:app\s*\(\s*(.*)\s*\)', '\\1', proc_name).strip().split(",")
        for a in str_args:
            if "=" in a:
                k,v = a.split("=")
                k = k.strip()
                v = v.strip()

                if v.lower() in ["true", "false"]:
                    v = v.lower() == "true"
                    if not v:
                        continue
                elif v[0] == '"':
                    v = v[1:-1]
                kwargs[k] = v

    from libretranslate.main import get_args
    sys.argv = ['--wsgi']

    for k in kwargs:
        ck = k.replace("_", "-")
        if isinstance(kwargs[k], bool) and kwargs[k]:
            sys.argv.append("--" + ck)
        else:
            sys.argv.append("--" + ck)
            sys.argv.append(kwargs[k])

    args = get_args()

    from libretranslate import flood, scheduler, secret, storage, cache
    storage.setup(args.shared_storage)
    cache.setup(args.translation_cache)
    scheduler.setup(args)
    flood.setup(args)
    secret.setup(args)